/*
  C extension to LDA-CCCP Numpy implementation
  Author: Jussi Kujala
  Thanks to ... for a useful example on C extensions with numpy arrays.
*/

#include "Python.h"
#include "arrayobject.h"
#include "C_lda.h"
#include <math.h>

/* the methods table  */
static PyMethodDef _C_ldaMethods[] = {
  {"lda_estep", lda_estep, METH_VARARGS},
  {"lda_lagrangians", lda_lagrangians, METH_VARARGS},
  {"lda_infer_new_doc", lda_infer_new_doc, METH_VARARGS},
  {NULL, NULL}     /* marks the end of this structure */
};

/* Initialize the C_lda functions */
// Module name must be _C_lda in compile and linked 
void init_C_lda()  {
  (void) Py_InitModule("_C_lda", _C_ldaMethods);
  import_array();  // Must be present for NumPy.  Called first after above line.
}

/*
  Input:
    -words
    -pwz input
    -pzd input
    -qwz output
    -qdz output
    -nwords
    -n_terms
    -n_docs
    -n_topics
    -smooth_fac
  Output:
    1 if succesfull
*/
static PyObject *lda_estep(PyObject *self, PyObject *args)
{
  int i,z;

  PyArrayObject* py_words;
  PyArrayObject *py_pwz,*py_pzd;
  PyArrayObject *py_qwz,*py_qzd; //should be zero, output is written to these
  int n_words,n_terms,n_docs,n_topics;
  double smooth_fac;

  /* Parse tuples separately since args will differ between C fcns */
  if (!PyArg_ParseTuple(args, "O!O!O!O!O!iiiid", 
      &PyArray_Type, &py_words,
      &PyArray_Type, &py_pwz,
      &PyArray_Type, &py_pzd,
      &PyArray_Type, &py_qwz,
      &PyArray_Type, &py_qzd,
      &n_words,
      &n_terms,
      &n_docs,
      &n_topics,
      &smooth_fac))  return NULL;
  if (NULL==py_words||NULL==py_pwz||NULL==py_pzd||NULL==py_qwz||NULL==py_qzd)  return NULL;
   
  //n=arrayin->dimensions[0];
  //m=arrayin->dimensions[1];
  //a=(double *) arrayin->data;   
  int n=py_pwz->dimensions[0];
  int m=py_pwz->dimensions[1];

  //printf("it works, i am in C!\n");
  //printf("parameters: n_words:%d n_terms:%d n_docs:%d n_topics:%d dim0:%d dim1:%d\n",n_words,n_terms,n_docs,n_topics,n,m);

  int* words = (int*) py_words->data;
  double* pwz = (double*) py_pwz->data;
  double* pzd = (double*) py_pzd->data;
  double* qwz = (double*) py_qwz->data;
  double* qzd = (double*) py_qzd->data;

  //printf("words[30:33]: %d %d %d\n",words[30],words[31],words[32]);

  int d;
  int w;
  int c;
  double* zvec = (double*) malloc(sizeof(double)*n_topics);
  if(NULL == zvec)
    return NULL;
  double s = 0.0;
  double mul = 0.0;
  double eps = smooth_fac/(n_terms*n_topics);
  for(i=0;i<n_words;i++) {
    d = words[3*i];
    w = words[3*i+1];
    c = words[3*i+2];
    s = 0.0;
    for(z=0;z<n_topics;z++) {
      zvec[z] = pwz[w*n_topics+z]*pzd[d*n_topics+z] + eps;
      s+=zvec[z];
    }
    if(s == 0.0)
      printf("sum=0 => floating point error will happen, this shouldn't happen, continuing still\n");
    mul = ((double)c)/s;
    for(z=0;z<n_topics;z++) {
      qwz[w*n_topics+z] += zvec[z]*mul;
      qzd[d*n_topics+z] += zvec[z]*mul;
    }
  }
  free(zvec);
  return Py_BuildValue("i", 1);
}


/*
  Input:
    -qwz 
    -qzd 
    -pwz0 
    -pdz0 
    -alfa
    -beta
    -n_terms
    -n_docs
    -n_topics
    -err
    -niter
    -maxdenom
  Output:
    1 if succesfull
*/  
static PyObject *lda_lagrangians(PyObject *self, PyObject *args) {
  PyArrayObject *py_qwz,*py_qzd;
  PyArrayObject *py_pwz0,*py_pzd0;
  int n_terms,n_docs,n_topics,n_iter;
  double alfa,beta,err,max_denom;

  /* Parse tuples separately since args will differ between C fcns */
  if (!PyArg_ParseTuple(args, "O!O!O!O!ddiiidid", 
      &PyArray_Type, &py_qwz,
      &PyArray_Type, &py_qzd,
      &PyArray_Type, &py_pwz0,
      &PyArray_Type, &py_pzd0,
      &alfa,
      &beta,
      &n_terms,
      &n_docs,
      &n_topics,
      &err,
      &n_iter,
      &max_denom))  return NULL;
  if (NULL==py_qwz||NULL==py_qzd||NULL==py_pwz0||NULL==py_pzd0)  return NULL;

  double* qwz = (double*) py_qwz->data;
  double* qzd = (double*) py_qzd->data;
  double* pwz0 = (double*) py_pwz0->data;
  double* pzd0 = (double*) py_pzd0->data;

  int z,d;

  double* q = (double*) malloc(sizeof(double)*n_terms);
  double* p0 = (double*) malloc(sizeof(double)*n_terms);

  double* b = (double*)malloc(sizeof(double)*n_terms);
  double* x = (double*)malloc(sizeof(double)*n_terms);

  if(NULL==q || NULL == p0 || NULL == b || NULL==x)
    return NULL;

  //go through topic distributions first
  for(z=0;z<n_topics;z++) {
    get_col(qwz,q,z,n_terms,n_topics); //these might be slow because mess around in memory
    get_col(pwz0,p0,z,n_terms,n_topics);
    normalize_q(q,p0,beta,n_terms,err,n_iter,b,x,max_denom);
    set_col(qwz,q,z,n_terms,n_topics);
    set_col(pwz0,p0,z,n_terms,n_topics);
  }
  free(q);
  free(p0);
  for(d=0;d<n_docs;d++) {
    normalize_q(&qzd[d*n_topics],&pzd0[d*n_topics],alfa,n_topics,err,n_iter,b,x,max_denom);    
  }
  free(b);
  free(x);
  return Py_BuildValue("i", 1);
}

void normalize_q(double* q, double* p0, double alfa, int n, double max_err,int max_iter, double* b, double* x, double max_denom) {

  if(sum(q,n)==0.0)
    return;

  max_denom = max_denom/n;

  int i;
  if(alfa==0.0) {
    for(i=0;i<n;i++)
      b[i]=0.0;
  } else {
    for(i=0;i<n;i++) {
      if(max_denom >= 0.0)
        p0[i] = (p0[i] < max_denom) ? max_denom : p0[i];
      b[i] = -alfa/p0[i];
      if(isnan(b[i]) || b[i] > 10000*n)
        b[i]=10000.0*n;
    }  
  }
  double mini = -min(b,n);
  double maxi = sum(q,n);
  double L = (maxi-mini)/2;
  double err = 1.0;
  int n_iter = 0;
  while(err >= max_err && n_iter < max_iter) {
    double s=0.0;
    for(i=0;i<n;i++) {
      x[i]=q[i]/(b[i]+L);
      s+=x[i];
    }
    err = 1.0-s;
    if(err > 0.0 && s > 0.0) {
      maxi=L;
      L = (L+mini)/2;
    } else {
      mini = L;
      L = (L+maxi)/2;
    }
    err = (err > 0.0) ? err : -err;
    n_iter+=1;
  }
  double s=0.0;
  for(i=0;i<n;i++) {
    q[i] = x[i];
    if(q[i] < 1e-40)
      q[i] = 0.0;
    s+=q[i];
  }
  for(i=0;i<n;i++)
    q[i] = q[i]/s;
}


/*
  Input:
    -words	:input data
    -pwz	:input model
    -pzd 	:output model
    -n_words 
    -n_topics
    -n_terms
    -n_max_iters 
    -min_diff
  Output:
    1 if succesfull
*/  

static PyObject *lda_infer_new_doc(PyObject *self, PyObject *args) {
  PyArrayObject *py_words,*py_pwz,*py_pzd;
  int n_words,n_topics,n_max_iters,n_terms;
  double min_diff;

  /* Parse tuples separately since args will differ between C fcns */
  if (!PyArg_ParseTuple(args, "O!O!O!iiiid", 
      &PyArray_Type, &py_words,
      &PyArray_Type, &py_pwz,
      &PyArray_Type, &py_pzd,
      &n_words,
      &n_topics,
      &n_terms,
      &n_max_iters,
      &min_diff))  return NULL;
  if (NULL==py_words||NULL==py_pwz||NULL==py_pzd)  return NULL;

  int* words = (int*) py_words->data;
  double* pwz = (double*) py_pwz->data;
  double* pzd = (double*) py_pzd->data;

  double* qzd = (double*) malloc(sizeof(double)*n_topics);
  double* zvec = (double*) malloc(sizeof(double)*n_topics);
  if(NULL==qzd||NULL==zvec)
    return NULL;

  int i,j,z;
  int d,w,c;
  double s,mul;
  double pzd_old,M,M_cur;
  for(i=0;i<n_max_iters;i++) {
    for(z=0;z<n_topics;z++)
      qzd[z] = 0.0;
    for(j=0;j<n_words;j++) {
      //d = words[3*j]; //discard
      w = words[3*j+1];
      c = words[3*j+2];
      s = 0.0;
      for(z=0;z<n_topics;z++) {
        zvec[z] = pwz[w*n_topics+z]*pzd[z];
        s+=zvec[z];
      }
      if(s!=0.0) {
        mul = ((double)c)/s;
        for(z=0;z<n_topics;z++) {
          qzd[z] += zvec[z]*mul;
        }
      }
    }
    s = 0.0;
    for(z=0;z<n_topics;z++){
      s+=qzd[z];
    }
    if(s==0.0) {
      printf("floating point error: sum qzd is zero\n");
      return Py_BuildValue("ddi",10e10,n_words*log(n_terms)/log(2),n_words);
    }
    M=1.0;
    for(z=0;z<n_topics;z++) {
      pzd_old = pzd[z];
      pzd[z] = qzd[z] / s;
      M_cur = my_abs(pzd[z]-pzd_old);
      M = (M < M_cur)? M_cur : M;
    }
    if(M<min_diff)
      break; //accuracy reached
    for(z=0;z<n_topics;z++) {
      if(pzd[z] < 1e-40)
        pzd[z] = 0.0;
    }
  }
  double real;
  double cap;
  int n_w;
  new_like(words, pwz, pzd, n_words, n_topics, n_terms,&real,&cap,&n_w);
  return Py_BuildValue("ddi",real,cap,n_w);
}

void new_like(int* words, double* pwz, double* pzd, int n_words, int n_topics, int max_w,double* r_real,double* r_cap,int* r_nw) {
  double real = 0.0;
  double cap = 0.0;
  int n_w = 0;
  double max_w_len = log(max_w)/log(2.0);
  int i,z,c;
  double ilog2 = 1.0/log(2);
  double s;
  for(i=0;i<n_words;i++) {
    s = 0.0;
    for(z=0;z<n_topics;z++)
      s += pwz[words[i*3+1]*n_topics+z]*pzd[z];
    if(s >1.0) {
      printf("error: pw was > 1");
      *r_real = -1;
      *r_cap = -1;
      *r_nw = -1;
      return;
    }
    if(s<=0.0)
      s = 10e10;
    else
      s = -ilog2*log(s);
    c = words[i*3+2];
    real += c*s;
    cap += (s > max_w_len) ? c*max_w_len : c*s;
    n_w+=c;
  }
  *r_real=real;
  *r_cap=cap;
  *r_nw=n_w;
}

double my_abs(double x){
  return (x>0)?x:-x;
}

double min(double* b,int n) {
  int i;
  double ret = b[0];
  for(i=0;i<n;i++)
    ret = (ret > b[i]) ? b[i] : ret;
  return ret;
}

double sum(double* x,int n) {
  int i;
  double ret = 0.0;
  for(i=0;i<n;i++)
    ret+=x[i];
  return ret;
}

//rows are continuous in memory
void get_col(double* pwz,double* p,int col,int n_y,int n_x) {
  int i;
  for(i=0;i<n_y;i++) {
    p[i] = pwz[i*n_x+col];
  }
}

void set_col(double* pwz,double* p,int col,int n_y,int n_x) {
  int i;
  for(i=0;i<n_y;i++) {
    pwz[i*n_x+col] = p[i];
  }
}

int my_isnan(double x) {
  return x!=x;
}

