#
#  lda-cccp : corpus -> topic_model
#
#  Copyright: Jussi Kujala (jussi.kujala@iki.fi)
#
#  
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#  This code is not perfect ;-)
#

description = """This program infers a document topic model for a given text corpus. Implementation is an EM algorithm for both PLSA objective and a kind of LDA objective. 
Corpus format is: a text file with three header lines (number of documents, number of terms, and number of unique word-document combinations) and lines of triples "docid wordid count". 
Supported options (only input must be specified):
--input string : Input corpus file to train from
--output string : Output file
--test string :Test model to test
--test_dumps string : Dumped models to test all
--test_res string : Output pickle of test result (test or test_dumps must be set)
--alfa float : Dirichlet parameter for topic distributions of documents
--beta float : Dirichlet parameter for topic distributions of documents
--n_topics integer : Number of topics
--n_cccp integer : Number CCCP steps
--n_em integer : Number of EM steps during one CCCP step
--n_plsi integer : Number of initial PLSA steps.
--new_doc_n_em_iters integer : Number of EM iters when inferring new documents.
--lin_on_edge : Linearize Dirichlet prior near edges.
--smooth_fac float : How strongly topic distributions are smoothed.
--init_model string : Initialise the model from the pickled file
--rewrite_ids : Rewrite document and word ids to form a continuous interval (might be buggy)
--dump yes/no : Whether intermediate results are dumped, default is no
--dump_freq int : How many iterations between dumping
--dump_fn string : Filename where to pickle intermediate results
--print_logprob : Print intermediate log-likelihood results.
--err_lagrangian_linesearch float : Error tolerance in linesearch for inferring Lagrangian for p(z|d) and p(w|z) distributions (1-err <= p <= 1+err)
--max_lagrangian_linesearch integer : Maximum number of steps of linesearch
"""
try:
  import _C_lda
  C_ext_loaded = True
except ImportError:
  C_ext_loaded = False 
  print "\n *** C extension not available, please compile, the Python version is slow and it might have bugs :-) ***\n"


import copy
import getopt
import logging
import os
import math
import numpy as np
import pickle
import cPickle
import random
import sys

logging.basicConfig(level=logging.DEBUG)

# enum of possible tasks
(ESTIMATE,INFER,TEST_DUMPS) = range(3)

# Contains all parameters and constants used in the algorithm
class Parameters:
  
  task = ESTIMATE

  max_w = 0 #number of words
  max_d = 0 #number of documents

  alfa = -0.5  #for topic distributions, really (\alpha-1)
  beta = -0.99  #for word distributions, really (\beta-1)
  n_topics = 100 #number of topics

  n_cccp = 60 #number of CCCP steps
  n_em = 1    #number of EM steps during each CCCP step
  n_plsi = 3  #number of EM steps in the initial PLSI phase

  rewrite_ids = False #rewrite doc and word ids to form a continuous interval

  final = False #is this round the final one (smaller error tolerances)

  err = 0.01  #error tolerance in the sum of probabilities when inferring probability distributions p(z|d) and p(w|z) (1-err <= p <= 1+err)
  max_iter = 30 #number of line search steps while inferring probability distributions
  final_err = 0.000001 #error during the final round

  no_zeros = False  #whether log of dirichlet is estimated with linear function at the edges; causes density to be finite
  lin_edge_on = 1.0 #if dirichlet linearized at edge, then exact position is this_param*(alfa+1)/(alfa+2)*topics)
  smooth_fac = 1.0 #smooth factor to topic distributions 

  new_doc_n_em_iters = 30 #how many iterations of EM algorithm when inferring p(z|d) distribution for new document
  new_doc_topic_accuracy = 0.00000001 #how much must at least each topic change during inferring p(z|d) for new document

  no_dumps = True #whether intermediate models are dumped
  dump_index = 0 #current iteration index
  dump_freq = 1 #how many iterations between not dumping the model to file
  dump_fn = "model_dump.pickle" #name for dumps files

  output_fn = None #final model output file name
  init_model_fn = None #used to initialise the model
  test_model_fn = None #model to test
  test_res_fn = None #file to dump test results to

  print_logprob = False #whether to print intermediate likelihood results

  c_extension =  C_ext_loaded #whether to use the C extension

class Corpus: #not really used
  fn = None
  max_w = None
  max_d = None
  words = None
  def __init__(self,fn,words):
    self.fn = fn
    self.words = words
    self.max_d = np.max(words[:,0]) +2 #good code sir
    self.max_w = np.max(words[:,1]) +2


class Model: # All information relating to a topic model over a corpus
  n_topics = None
  pzd = None #topic-document -distributions: p[d,:] is one
  pwz = None #word-topic -distributions: p[:,z] is one
  doc_map = None  #if ids of documents are mapped to new values, then this shows how it was done
  word_map = None #same for word ids
  def __init__(self,pzd,pwz):
    self.pzd = pzd
    self.pwz = pwz
    #assert pzd.shape[1] == pwz.shape[1]
    self.n_topics = pwz.shape[1]


# solves the lagrangian of the optimization task in the EM-algorithm
# is a line search algorithm
# param q : q-values corresponding to either (document,topic) or (topic,word) pair; is an unnormalized probability distribution
# param p0 : previous probability values
# param alfa : dirichlet prior paramater
# param params : contains the above parameter class
# output: normalized probability distribution, where result[i] = q[i] / (-alfa/p0[i]+L) and L is the Lagrangian that is solved
def solve_L(q,p0,alfa,params):
  l = len(p0)
  assert l > 1
  if params.no_zeros: # estimate with a linear function at the edges
    max_denom = params.lin_edge_on*(1.0-(float(params.dump_index+2-params.n_plsi))/(params.n_cccp+1))
    p0 = np.array([max(val,max_denom/l) for val in p0])

  # we solve a vector x = a/(b+L) so that sum(x)=1
  b = None
  if alfa == 0.0:
    b = np.zeros((len(p0),),dtype=float)
  else:
    b = -alfa/p0
  mini = -np.min(b)
  maxi = np.sum(q)
  assert maxi < mini
  L = (maxi-mini)/2.0
  err = 1.0
  x = q
  n_iter = 0
  s = 0.0
  err_tolerance = params.err
  if params.final:
    err_tolerance = params.final_err
  while err > err_tolerance and n_iter < params.max_iter:
    x = q / (b+L)
    s = np.sum(x)
    err = 1.0-s
    if err > 0.0 and s > 0.0:
      maxi = L
      L = (L+mini)/2
    else:
      mini = L
      L = (L+maxi)/2
    err = np.abs(err)
    n_iter = n_iter + 1
  return x   

#wrapper that calls C extension
def py_wrapper_to_C_estep(words,pwz,pzd,qwz,qzd,n_terms,n_docs,n_topics,params):
  n_words = words.shape[0]
  smooth_fac = params.smooth_fac
  if params.final or params.dump_index < params.n_plsi:
    smooth_fac = 0.0
  type_test = type(np.zeros(2,dtype=float))
  assert type(words)==type_test
  assert type(pwz)==type_test
  assert type(pzd)==type_test
  assert type(qwz)==type_test
  assert type(qzd)==type_test
  assert n_terms > 0
  assert n_docs > 0
  assert n_topics > 0
  #t1 = time.time()
  ret = _C_lda.lda_estep(words,pwz,pzd,qwz,qzd,n_words,n_terms,n_docs,n_topics,smooth_fac)
  #t2 = time.time()
  #print 'Estep took %0.3f ms' % ((t2-t1)*1000.0)
  assert ret == 1

#wrapper that calls C extension
def py_wrapper_to_C_lagrangians(qwz,qzd,cccp_model,params):
  err = params.err
  if params.final:
    err=params.final_err
  n_docs = params.max_d
  n_terms = params.max_w
  n_topics = params.n_topics
  assert n_topics == qwz.shape[1]
  assert n_topics == qzd.shape[1]
  alfa = params.alfa
  beta = params.beta
  pwz0 = cccp_model.pwz
  pzd0 = cccp_model.pzd
  assert n_topics == pwz0.shape[1]
  assert n_topics == pzd0.shape[1]
  max_denom = -1.0
  if params.no_zeros:
    max_denom = params.lin_edge_on*(1.0-(float(params.dump_index+2-params.n_plsi))/(params.n_cccp+1))
    if max_denom > 1.0:
      max_denom = 1.0
  type_test = type(np.zeros(2,dtype=float))
  assert type(qwz)==type_test
  assert type(qzd)==type_test
  assert type(pwz0)==type_test
  assert type(pzd0)==type_test
  assert n_terms > 0
  assert n_docs > 0
  assert n_topics > 0
  ret = _C_lda.lda_lagrangians(qwz,qzd,pwz0,pzd0,alfa,beta,n_terms,n_docs,n_topics,err,params.max_iter,max_denom)
  assert ret == 1

#wrapper that calls C extension
def py_wrapper_to_C_infer_new_doc(words,pwz,pzd,params):
  n_words = words.shape[0]
  n_topics = params.n_topics
  type_test = type(np.zeros(2,dtype=float))
  n_max_iters = params.new_doc_n_em_iters
  min_diff = params.new_doc_topic_accuracy
  n_terms = params.max_w
  assert type(words)==type_test
  assert type(pwz)==type_test
  assert type(pzd)==type_test
  assert n_words > 0
  assert n_topics > 0
  ret = _C_lda.lda_infer_new_doc(words,pwz,pzd,n_words,n_topics,n_terms,n_max_iters,min_diff)
  assert ret != 0
  assert type(ret)==type(tuple())
  assert len(ret)==3
  return ret

#E-step of the algorithm in Python
def py_estep(words,pwz,pzd,qwz,qzd):
  #go through each word in the corpus, update all q-distributions based on this knowledge
  zs = pwz[0,:]
  for i,(d,w,c) in enumerate(words): #(docid,wordid,count)
    if i % 100000 ==0:
      logging.info("lda_em_iter %d, word %d" %(i,w))
    zs[:] = pwz[w,:]*pzd[d,:]
    zs *= np.float(c)/zs.sum()
    assert zs.sum > 0.0
    qwz[w,:] += zs
    qzd[d,:] += zs

#M-step of the algorithm in Python
def py_lagrangians(qwz,qzd,cccp_model,params):
  for z in range(0,params.n_topics):
    if np.sum(qwz[:,z]) == 0.0:
      continue
    qwz[:,z] = solve_L(qwz[:,z],cccp_model.pwz[:,z],params.beta,params)
    assert np.sum(qwz[:,z]) > 1-params.err and np.sum(qwz[:,z]) < 1+params.err
  for d in range(1,params.max_d):
    if np.sum(qzd[d,:])==0:
      continue
    qzd[d,:] = solve_L(qzd[d,:],cccp_model.pzd[d,:],params.alfa,params)
    assert np.sum(qzd[d,:]) > 1-params.err and np.sum(qzd[d,:]) < 1+params.err

# performs one step of EM-algorithm
# param words : list of (docid,wordid,count)
# param old_model : p[0] = p(z|d), p[1] = p[w|z]
# param p0 : previous estimates
def lda_em_iter(words,current_model,cccp_model,params):
  #all the magic happens here
  pzd = current_model.pzd
  pwz = current_model.pwz
  qzd = np.zeros((params.max_d,params.n_topics),dtype=float)
  qwz = np.zeros((params.max_w,params.n_topics),dtype=float)
  if params.c_extension:
    py_wrapper_to_C_estep(words,pwz,pzd,qwz,qzd,params.max_w,params.max_d,params.n_topics,params)
  else:
    py_estep(words,pwz,pzd,qwz,qzd)       
  #logging.info("solving lagrangians...")
  if params.c_extension:
    py_wrapper_to_C_lagrangians(qwz,qzd,cccp_model,params)
  else:
    py_lagrangians(qwz,qzd,cccp_model,params)
  q = Model(pzd=qzd,pwz=qwz)
  return q

# calls lda_em_iter for n_em iterations
# params : see previous function
def lda_em(words,current_model,cccp_model,params,n_em):
  for i in range(0,n_em):
    if params.print_logprob:
      logging.info("EM-iter: %d, %s"  %(i+1,str_logprob(words,current_model,params)))
    current_model = lda_em_iter(words,current_model,cccp_model,params)
    smart_dump(current_model,params)
  return current_model

# helper function to initialise a probability matrix : rows are random probability distributions
def init_arr_rows(x,y):
  a = np.zeros((x,y),dtype=float)
  for r in range(0,x):
    a[r,:] = np.random.random_sample((y,))
    a[r,:] = a[r,:] / sum(a[r,:])
  return a

# same as previous function, but this time columns are random probability distributions
def init_arr_cols(x,y):
  a = np.zeros((x,y),dtype=float)
  for c in range(0,y):
    a[:,c] = np.random.random_sample((x,))
    a[:,c] = a[:,c] / sum(a[:,c])
  return a

# initialises initial parameters randomly : in practise initialising to random smoothed documents in the corpus would probably work better
def init_model(words,max_d,max_w,n_topics):
  #arrays for each document and for each topic
  pzd = init_arr_rows(max_d,n_topics)
  pwz = init_arr_cols(max_w,n_topics)
  return Model(pzd=pzd,pwz=pwz)

# initialise parameters to uniform values (used in PLSI to estimate the initial flat distribution)
def init_uni_model(max_d,max_w,n_topics):
  pzd = np.ones((max_d,n_topics),dtype=float)
  pwz = np.ones((max_w,n_topics),dtype=float)
  pzd = pzd / n_topics
  pwz = pwz / max_w
  return Model(pzd=pzd,pwz=pwz)

# CCCP algorithm for LDA
def lda_cccp(words,params,prev_model=None):
  logging.info("called lda-cccp")
  model = None
  if prev_model != None:
    model = prev_model
  else:
    model = init_model(words,params.max_d,params.max_w,params.n_topics)
  model_uni = init_uni_model(params.max_d,params.max_w,params.n_topics) #the dirichlet is estimated at uniform first (=zero change) 
  logging.info("solving initial PLSI")
  #smart_dump(model,params,force_dump=True)
  assert model is not None
  model=lda_em(words,model,model_uni,params,params.n_plsi) #this is in effect emulation of PLSI, replace with a real one if you have time
  del model_uni
  #smart_dump(model,params,force_dump=True)
  cccp_model = copy.deepcopy(model)
  logging.info("CCCP iterations...")
  for i in range(0,params.n_cccp):
    logging.info("CCCP iter: %d" %(i+1))
    if i == params.n_cccp-1:
      params.final = True
    model=lda_em(words,model,cccp_model,params,params.n_em)
    cccp_model = copy.deepcopy(model)
  if params.print_logprob:
    logging.info("final complexity: %s"  %(str_logprob(words,model,params)))
  return model

# returns log probability of the data and model parameters given the model
def logprob(words,model,params):
  pzd = model.pzd
  pwz = model.pwz
  lg_pr = 0.0
  cap_log_pr = 0.0
  pw = 0
  n_words = 0
  lpw = np.float(0.0)
  max_w_len = np.log2(params.max_w)
  for (d,w,c) in words:
    lpw = np.log2(np.sum(pwz[w,:]*pzd[d,:]))
    lg_pr += -c*lpw
    cap_log_pr += c*min(-lpw,max_w_len)
    n_words +=c
  lg_data_pr=lg_pr
  lg_pr = 0.0  
  for d in range(0,params.max_d):
    for z in range(0,params.n_topics):
      if pzd[d,z] == 0:
        continue
      lg_pr += -params.alfa*np.log2(pzd[d,z])
  for z in range(0,params.n_topics):
    #(doc_id_map,pzd,likes)
    for w in range(0,params.max_w):
      if pwz[w,z] == 0:
        continue
      lg_pr += -params.beta*np.log2(pwz[w,z])
  lg_model_pr = lg_pr
  return (lg_data_pr,lg_model_pr,cap_log_pr,n_words)

#return log-probability in readable format
def str_logprob(words,model,params):
  logging.info("inferring log-prob")
  (d_pr,m_pr,capped_pr,n_words) = logprob(words,model,params)
  return "log-probability: %d, data: %d, model: %d, impr: %f, capped_impr: %f" %(int(d_pr+m_pr),int(d_pr),int(m_pr),float(d_pr)/(n_words*np.log2(params.max_w)),float(capped_pr)/(n_words*np.log2(params.max_w)))

#computes log-likelihood for new document given its pzd
#returns two values (real_lik,mdl_lik) where first is computed from real probabilities and second where low probability effect is capped
def new_lik(words,pzd,pwz,max_w):
  ret = dict()
  real = 0.0
  cap = 0.0
  n_w = 0
  max_w_len = np.log2(max_w)
  x=0.0
  for i in range(words.shape[0]):
    x = -np.log2(np.sum(pwz[words[i,1],:]*pzd))
    if np.isnan(x):
      x = np.inf
    if x < 0.0:
      print x, w,pzd, pwz[w,:]
      assert False
    real += words[i,2]*x
    cap += np.min((words[i,2]*x,words[i,2]*max_w_len))
    n_w += words[i,2]
  return (real,cap,n_w)

# aggregates likelihoods together
# input:
#  likes = dict from docid to dict containing keys "real" "cap" and "n_w"
#  n_words = number of distinct words in the corpus
# output: (lik,cap,n_w,impr):
#  lik = corpus likelihood
#  cap = capped likelihood
#  n_w = total number of words in the corpus
#  impr = improvement ratio compared to flat coding of words
def get_total_new_like(likes,max_w):
  real = 0.0
  cap = 0.0
  n_w = 0
  for d_k in likes:
    (dreal,dcap,dwn) = likes[d_k]
    real += dreal
    cap += dcap
    n_w += dwn
  impr = cap/(n_w*np.log2(max_w))
  return (real,cap,n_w,impr)

# processes new document
def infer_new_doc(words,pwz,params):
  n_topics = pwz.shape[1]
  pdz = np.random.rand(n_topics)
  pdz = pdz/np.sum(pdz)
  err = 0
  for i in range(params.new_doc_n_em_iters):
    qzd = np.zeros(n_topics,dtype=float)
    t = np.zeros(n_topics,dtype=float)
    s = np.float(0.0)
    for i in range(words.shape[0]):
      d=words[i,0]
      w=words[i,1]
      c=words[i,2]
      t = pwz[w,:]*pdz
      s = np.sum(t)
      if s != 0.0:
        qzd += c*t / s 
    pdz_old = pdz
    pdz = qzd / np.sum(qzd)
    if np.sum(np.isnan(pdz)) > 0:
      print pdz, pdz_old, qzd
      assert False
    max_err = np.max(np.abs(pdz_old - pdz))
    if max_err < params.new_doc_topic_accuracy:
      break
  pdz = pdz /np.sum(pdz)
  return pdz

# given a model, infer topic distributions for completely new documents 
def infer_new_docs(words,pwz,params,preprocessed_words=None):
  logging.info("inferring p(z|d) distributions for new documents")
  max_w = params.max_w
  n_topics = pwz.shape[1]
  assert params.n_topics == n_topics
  docs = None
  n_docs = 0
  if preprocessed_words==None:
    doc_counts = dict()
    for i in range(words.shape[0]): #collect all doc_ids
      d=words[i,0]
      if doc_counts.has_key(d):
        doc_counts[d]+=1
      else:
        doc_counts[d]=1
    n_docs = len(doc_counts.keys()) 
    docs = dict()
    for doc_id in doc_counts:
      docs[doc_id] = np.zeros((doc_counts[doc_id],3),dtype=int)
      doc_counts[doc_id]=0
    logging.info("splitting data to separate documents...")
    j=-1
    for i in range(words.shape[0]): #find words in one doc
      j = doc_counts[words[i,0]]
      docs[words[i,0]][j,0] = words[i,0]
      docs[words[i,0]][j,1] = words[i,1]
      docs[words[i,0]][j,2] = words[i,2]
      doc_counts[words[i,0]]+=1
    logging.info("splitted")
  else:
    docs = preprocessed_words
    n_docs = len(docs)
  pzd = np.zeros((n_docs,n_topics),dtype=float) 
  doc_id_map = dict()
  likes = dict()
  for i , doc_id in enumerate(docs): #find pdz for all docs
    if i %1000 == 0:
      logging.info("inferring on doc:%d" %i)
    doc_id_map[("o2n",doc_id)] = i
    doc_id_map[("n2o",i)] = doc_id
    if params.c_extension:
      pzd_t = np.random.rand(n_topics)
      pzd_t = pzd_t/np.sum(pzd_t)
      likes[doc_id] = py_wrapper_to_C_infer_new_doc(docs[doc_id],pwz,pzd_t,params) # returns (like,cap_like,n_w)
      pzd[i,:] = pzd_t
    else:
      pzd[i,:] = infer_new_doc(docs[doc_id],pwz,params)
      likes[doc_id] = new_lik(docs[doc_id],pzd[i,:],pwz,max_w)
  total_like = get_total_new_like(likes,max_w) 
  return (doc_id_map,pzd,likes,total_like)

# returns a new corpus where doc ids contain no gaps
# i.e. ids are rewritten to be continous ranges
# output:
#  words: new corpus
#  doc_map: mappings between new and old document ids doc_map[("o2n,old_id)] = new_id, doc_map[("n2o,new_id)] = old_id
#  word_map: mappings between new and old word ids
def rewrite_compact_ids(words):
  logging.info("rewriting the ids of the words...")
  doc_ids = set()
  word_ids = set()
  for (d,w,c) in words: #collect all doc_ids
    doc_ids.add(d) 
    word_ids.add(w)
  doc_ids = list(doc_ids)
  doc_ids.sort()
  word_ids = list(word_ids)
  word_ids.sort()
  doc_map = dict()
  word_map = dict()
  for i,id_ in enumerate(doc_ids):
    doc_map[("n2o",i+1)]=id_    #new to old: n2o
    doc_map[("o2n",id_)] = i+1  #old to new: o2n
  for i,id_ in enumerate(word_ids):
    word_map[("n2o",i+1)]=id_
    word_map[("o2n",id_)] = i+1
  for i,(d,w,c) in enumerate(words):
    words[i,0] = doc_map[("o2n",d)]
    words[i,1] = word_map[("o2n",w)]
  logging.info("rewritten")
  return (words,doc_map,word_map)

#maps ids to new ids
def rewrite_ids(data,id_map,col,rw_type):
  for i in range(data.shape[0]):
    key = (rw_type,data[i,col])
    if not id_map.has_key(key):
      data[i,col] = 0 # unknown word
    else:  
      data[i,col] = id_map[key]
  return data

# split data to a training and test set
# param f_test : how many percent of docs in test
def split_data(words,f_test,max_d):
  n_test = int(max_d * f_test)
  ids_te = [False] * max_d
  ids_te[0:n_test] = [True] * n_test
  random.shuffle(ids_te)
  n_te = 0
  n_tr = 0
  for (d,w,c) in words:
    if ids_te[d]:
      n_te +=1
    else:
      n_tr +=1
  words_te = np.zeros((n_te,3),dtype=int)
  words_tr = np.zeros((n_tr,3),dtype=int)
  i_te = 0
  i_tr = 0
  for (d,w,c) in words:
    if ids_te[d]:
      words_te[i_te,:] = np.array((d,w,c))
      i_te +=1
    else:
      words_tr[i_tr,:] = np.array((d,w,c))
      i_tr +=1
  assert i_te == n_te and i_tr == n_tr
  return (words_tr,words_te)

#parse pwz from file
def read_pwz(fn):
  f=open(fn,"r")
  pwz=pickle.load(f)
  f.close()
  return pwz

#read input corpus
def read_corpus(fn):
  logging.info("reading the corpus from file %s..." %fn)
  f = open(fn,"r")
  ndocs = int(f.readline())
  nterms = int(f.readline())
  nwords = int(f.readline())
  words = np.zeros((nwords,3),dtype=int)
  i=0
  for line in f:
    if i%50000==0:
      logging.info("read line %d" %i)
    d,w,c = line.split()
    words[i,0] = int(d)
    words[i,1] = int(w)
    words[i,2] = int(c)
    i+=1
  f.close()
  logging.info("read the corpus")
  corpus = Corpus(fn,words)
  return corpus

#write model
def dump_model(model,filename):
  f = open(filename,"w")
  cPickle.dump(model,f,protocol=cPickle.HIGHEST_PROTOCOL)
  f.close()

g_doc_map = None #good code :)
g_word_map = None
# dumps the model on the fly for later processing
def smart_dump(model,params,force_dump=False):
  params.dump_index += 1
  if params.no_dumps:
    return
  if not force_dump and params.dump_index % params.dump_freq != 0:
    return
  global g_doc_map
  global g_word_map
  fn = "%s.%d" %(params.dump_fn,params.dump_index)
  model.word_map = g_word_map
  model.doc_map = g_doc_map
  f= open(fn,"wb")
  cPickle.dump(model,f,protocol=cPickle.HIGHEST_PROTOCOL)
  f.close()
  model.word_map=None
  model.doc_map=None


#code given below is UI-specific
def preprocess(data):
  logging.info("preprocessing...")
  data = data[np.argsort(data[:,1]),:]
  prev_ind = 0
  prev_word = data[0,1]
  for i,(d,w,c) in enumerate(data):
    if w == prev_word:
      continue
    data[prev_ind:i,:] = data[prev_ind:i,:][np.argsort(data[prev_ind:i,0]),:]
    prev_word = w
    prev_ind = i
  data[prev_ind:,:] = data[prev_ind:,:][np.argsort(data[prev_ind:,0]),:]
  logging.info("preprocessing done")
  return data

def usage():
  print description

def parse_cp():
  try:
    opts, args = getopt.getopt(sys.argv[1:], "x", ["help",
                                                  "input=", 
                                                  "output=", 
                                                  "test=", 
                                                  "test_dumps=",
                                                  "test_res=",
                                                  "alfa=", 
                                                  "beta=", 
                                                  "n_topics=", 
                                                  "n_cccp=", 
                                                  "n_em=", 
                                                  "n_plsi=",
                                                  "lin_on_edge",
                                                  "smooth_fac=",
                                                  "new_doc_n_em_iters=", 
                                                  "init_model=",
                                                  "rewrite_ids",
                                                  "dump=",
                                                  "dump_freq=",
                                                  "dump_fn=",
                                                  "print_logprob",
                                                  "max_convergence=", 
                                                  "err_lagrangian_linesearch=", 
                                                  "max_lagrangian_linesearch=",
                                                  "c_extension="])
  except getopt.GetoptError, err:
    # print help information and exit:
    print str(err) # will print something like "option -a not recognized"
    usage()
    sys.exit(2)
  params = Parameters()
  inputf = None
  for o, a in opts:
    if o in ("-h", "--help"):
      usage()
      sys.exit(0)
    elif o in ("--input"):
      inputf = a
    elif o in ("--output"):
      params.output_fn = a
    elif o in ("--test"):      
      params.task = INFER
      params.test_model_fn = a
    elif o in ("--test_dumps"):
      params.task = TEST_DUMPS
      params.test_model_fn = a
    elif o in ("--test_res"):
      params.test_res_fn = a
    elif o in ("--alfa"):
      params.alfa = float(a)-1.0
    elif o in ("--beta"):
      params.beta = float(a)-1.0
    elif o in ("--n_topics"):
      params.n_topics = int(a)
    elif o in ("--n_cccp"):
      params.n_cccp = int(a)
    elif o in ("--n_em"):
      params.n_em = int(a)
    elif o in ("--n_plsi"):
      params.n_plsi = int(a)
    elif o in ("--lin_on_edge"):
      params.no_zeros = True
    elif o in ("--smooth_fac"):
      params.smooth_fac = float(a)
    elif o in ("--new_doc_n_em_iters"):
      params.new_doc_n_em_iters = int(a)
    elif o in ("--init_model"):
      params.init_model_fn = a
    elif o in ("--rewrite_ids"):
      params.rewrite_ids = True
    elif o in ("--dump"):
      if a == "yes":
        params.no_dumps = False
      elif a == "no":
        params.no_dumps = True
      else:
        print "dump argument must be [yes|no]"
        exit(-2)
    elif o in ("--dump_freq"):
      params.dump_freq = int(a)
    elif o in ("--dump_fn"):
      params.dump_fn = a
    elif o in ("--print_logprob"):
      params.print_logprob = True
    elif o in ("--max_convergence"):
      print "not supported yet"
      exit(-2)
    elif o in ("--err_lagrangian_linesearch"):
      params.err = float(a)
    elif o in ("--max_lagrangian_linesearch"):
      params.max_iter = int(a)
    elif o in ("--c_extension"):
      params.c_extension = bool(int(a))
    else:
      assert False, "unhandled option"
  if inputf == None:
    usage()
    sys.exit(-2)
  return (inputf,params)

def main():
  (inputf,params) = parse_cp()
  assert params.alfa <= 0.0 and params.beta <= 0.0
  corpus = read_corpus(inputf)
  data = corpus.words
  doc_map = None
  word_map = None
  if params.rewrite_ids:
    assert params.task == ESTIMATE #not implemented otherwise
    (data,doc_map,word_map) = rewrite_compact_ids(data)
    global g_doc_map
    global g_word_map
    g_doc_map = doc_map
    g_word_map = word_map
  params.max_d = np.max(data[:,0]) +2 #good code sir
  params.max_w = np.max(data[:,1]) +2 #still good code sir!
  if params.task == ESTIMATE:
    data = preprocess(data)
    init_model = None
    if params.init_model_fn is not None:
      init_model = cPickle.load(open(params.init_model_fn,"r"))
      logging.info("loaded previous model, note that doc ids and word ids should match (no incompatible rewrites)")
    model = lda_cccp(data,params,init_model) #NCCCP,NEM,NPLSI
    del data
    smart_dump(model,params,True)
    model.doc_map = doc_map
    model.word_map = word_map
    if params.output_fn is not None:
      f=open(params.output_fn,"wb")
      cPickle.dump(model,f,protocol=cPickle.HIGHEST_PROTOCOL)
      f.close()
  elif params.task == INFER:
    test_model = pickle.load(open(params.test_model_fn,"r"))
    pwz = test_model.pwz
    params.max_w = pwz.shape[0]+2
    params.n_topics = pwz.shape[1]
    print "max_w:%d n_topics:%d" %(params.max_w,params.n_topics)
    if test_model.word_map is not None:
      params.max_w = max(params.max_w,max(test_model.word_map.values())+2)
      if word_map is not None:
        data = rewrite_ids(data,word_map,col=1,rw_type="n2o")
      data = rewrite_ids(data,test_model.word_map,col=1,rw_type="o2n")
    del test_model
    for i in range(data.shape[0]): #remove unseen words to prevent indexing errors
      w=data[i,1]
      if w > pwz.shape[0]-1:
        #logging.info("unknown word:%d"%w)
        data[i,1] = 0
    (doc_id_map,pzd,likes,like) = infer_new_docs(data,pwz,params)
    logging.info("(test set: number of words: %d log-like: %d log-like/uniform: %f)" %(like[1],like[2],like[3]))
    #print like
    #print params.max_w
    if params.output_fn is not None:
      test_model.doc_map = doc_id_map
      #test_model.word_map = word_map
      f=open(params.output_fn,"wb")
      cPickle.dump(test_model,f)
      f.close()
    if params.test_res_fn is not None:
      f=open(params.test_res_fn,"w")
      cPickle.dump(like,f)
      f.close()
  elif params.task == TEST_DUMPS:
    dump_models = pickle.load(open(params.test_model_fn,"r"))
    assert len(dump_models) > 1
    tmp_model = dump_models[-1]
    if tmp_model.word_map is not None:
      if word_map is not None:
        data = rewrite_ids(data,word_map,col=1,rw_type="n2o")
      data = rewrite_ids(data,tmp_model.word_map,col=1,rw_type="o2n")
    l = []
    for i in range(1,len(dump_models)):
      model = dump_models[i]
      pwz = model.pwz
      (doc_id_map,pzd,likes,like) = infer_new_docs(data,pwz,params)
      l.append(like)
    if params.test_res_fn is not None:
      f=open(params.test_res_fn,"w")
      pickle.dump(l,f)
      f.close()
  else:
    assert False


if __name__ == "__main__":
  main()


